#include <errno.h>
#include <stdlib.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>

void __attribute__ ((constructor)) my_init(void)
{
}

void __attribute__ ((destructor)) my_fini(void)
{
}

#if PUT_BYTES == 1
static void put_random_bytes(char *dummy, int n)
{
	int fd = open(DEV_RANDOM, O_WRONLY);
	if (fd == -1)
	{
		syslog(LOG_EMERG, "libprngwrap: failed to open " DEV_RANDOM " (write)! Reason: %m");
		return;
	}

	while(n > 0)
	{
		int rc = write(fd, dummy, n);

		if (rc == 0)
		{
			syslog(LOG_EMERG, "libprngwrap: got EOF while writing to " DEV_RANDOM "!");
			break;
		}
		else if (rc < 0)
		{
			if (errno == EINTR || errno == EAGAIN)
				continue;

			syslog(LOG_EMERG, "libprngwrap: error while writing to " DEV_RANDOM "! Reason: %m");
			break;
		}

		dummy += rc;
		n -= rc;
	}

	if (close(fd) == -1)
		syslog(LOG_EMERG, "libprngwrap: error while closing fd to " DEV_RANDOM "! Reason: %m");
}
#endif

static void get_random_bytes(char *dummy, int n)
{
	int fd = open(DEV_RANDOM, O_RDONLY);
	if (fd == -1)
	{
		/* if /dev/urandom is not available, use lrand48() to fill the buffer
		 * first initialize its seed with some random data
		 */
#if NODEV_PRNG == 1
		struct timeval tv;
		static char shown_warning = 0;

		if (!shown_warning)
		{
			syslog(LOG_EMERG, "libprngwrap: failed to open " DEV_RANDOM " (read), using PRNG!");
			shown_warning = 1;
		}

		if (gettimeofday(&tv, NULL) == -1)
			syslog(LOG_EMERG, "libprngwrap: gettimeofday() failed! Reason: %m");
		else 
		{
			int index, nbytes = sizeof(long int);
			long int seed;
			for(index=0; index<n; index++)
				((char *)&seed)[index % nbytes] ^= dummy[index];
			seed ^= tv.tv_sec;
			seed ^= tv.tv_usec;
			seed ^= (long int)dummy;
			srand48(seed);

			for(index=0; index<n; index++)
				dummy[index] = lrand48() & 0xff;
		}
#else
		syslog(LOG_EMERG, "libprngwrap: failed to open " DEV_RANDOM " (read)! Reason: %m");
		exit(1);
#endif
	}
	else
	{
		while(n > 0)
		{
			int rc = read(fd, dummy, n);

			if (rc == 0)
			{
				syslog(LOG_EMERG, "libprngwrap: got EOF while reading from " DEV_RANDOM "!");
				exit(1);
			}
			else if (rc < 0)
			{
				if (errno == EINTR || errno == EAGAIN)
					continue;

				syslog(LOG_EMERG, "libprngwrap: error while reading from " DEV_RANDOM "! Reason: %m");
				exit(1);
			}

			dummy += rc;
			n -= rc;
		}

		if (close(fd) == -1)
			syslog(LOG_EMERG, "libprngwrap: error while closing fd to " DEV_RANDOM "! Reason: %m");
	}
}

int rand(void)
{
	int dummy;

	get_random_bytes((char *)&dummy, sizeof(dummy));

	return abs(dummy) % RAND_MAX;
}

int rand_r(unsigned int *seedp)
{
	int dummy;

	get_random_bytes((char *)&dummy, sizeof(dummy));

	return abs(dummy) % RAND_MAX;
}

void srand(unsigned int seed)
{
#if PUT_BYTES == 1
	put_random_bytes((char *)&seed, sizeof(seed));
#endif
}

/* --- */

long int random(void)
{
	long int dummy;

	get_random_bytes((char *)&dummy, sizeof(dummy));

	return abs(dummy) % RAND_MAX;
}

void srandom(unsigned int seed)
{
#if PUT_BYTES == 1
	put_random_bytes((char *)&seed, sizeof(seed));
#endif
}

char *initstate(unsigned int seed, char *state, size_t n)
{
	static char previous_array[256];

#if PUT_BYTES == 1
	put_random_bytes(state, n);
	put_random_bytes((char *)&seed, sizeof(seed));
#endif

	return previous_array;
}

char *setstate(char *state)
{
	static char previous_array[256];

	return previous_array;
}

/* --- */

static double double_rand()
{
	unsigned long long int v1, v2;

	do
	{
		get_random_bytes((char *)&v1, sizeof(v1));
		get_random_bytes((char *)&v2, sizeof(v2));
	}
	while(v1 == v2);

	if (v1 > v2)
		return (double)v2 / (double)v1;
	else
		return (double)v1 / (double)v2;
}

double drand48(void)
{
	return double_rand();
}

double erand48(unsigned short xsubi[3])
{
	return double_rand();
}

long int lrand48(void)
{
	long int dummy;

	get_random_bytes((char *)&dummy, sizeof(dummy));
	dummy = abs(dummy);

	return dummy & 0xffffffff;
}

long int nrand48(unsigned short xsubi[3])
{
	long int dummy;

	get_random_bytes((char *)&dummy, sizeof(dummy));
	dummy = abs(dummy);

	return dummy & 0xffffffff;
}

long int mrand48(void)
{
	int dummy;

	get_random_bytes((char *)&dummy, sizeof(dummy));

	return dummy & 0xffffffff;
}

long int jrand48(unsigned short xsubi[3])
{
	int dummy;

	get_random_bytes((char *)&dummy, sizeof(dummy));

	return dummy & 0xffffffff;
}

void srand48(long int seed)
{
#if PUT_BYTES == 1
	put_random_bytes((char *)&seed, sizeof(seed));
#endif
}

unsigned short *seed48(unsigned short seed16v[3])
{
	static unsigned short iseed16v[3] = { 0 };

#if PUT_BYTES == 1
	put_random_bytes((char *)&(seed16v[0]), sizeof(unsigned short));
	put_random_bytes((char *)&(seed16v[1]), sizeof(unsigned short));
	put_random_bytes((char *)&(seed16v[2]), sizeof(unsigned short));
#endif

	return iseed16v;
}

void lcong48(unsigned short param[7])
{
#if PUT_BYTES == 1
	int index;
	for(index=0; index<7; index++)
		put_random_bytes((char *)&param[index], sizeof(unsigned short));
#endif
}

int random_r(struct random_data *buf, int32_t *result)
{
	if (result)
	{
		get_random_bytes((char *)result, sizeof(int32_t));

		return 0;
	}

	return -1;
}

int srandom_r(unsigned int seed, struct random_data *buf)
{
#if PUT_BYTES == 1
	put_random_bytes((char *)&seed, sizeof(seed));
#endif

	return 0;
}

int initstate_r(unsigned int seed, char *statebuf, size_t statelen, struct random_data *buf)
{
	if (statebuf)
	{
#if PUT_BYTES == 1
		put_random_bytes(statebuf, statelen);
		put_random_bytes((char *)&seed, sizeof(seed));
#endif
initstate(seed, statebuf, statelen);
		return 0;
	}

	return -1;
}

int setstate_r(char *statebuf, struct random_data *buf)
{
	if (statebuf)
		return 0;

	return -1;
}

int drand48_r(struct drand48_data *buffer, double *result)
{
	if (result)
	{
		*result = double_rand();
		return 0;
	}

	return -1;
}

int erand48_r(unsigned short xsubi[3], struct drand48_data *buffer, double *result)
{
	if (result)
	{
		*result = double_rand();
		return 0;
	}

	return -1;
}

int lrand48_r(struct drand48_data *buffer, long int *result)
{
	if (result)
	{
		get_random_bytes((char*)result, sizeof(long int));
		*result = abs(*result);
		(*result) &= 0xffffffff;
		return 0;
	}

	return -1;
}

int nrand48_r(unsigned short int xsubi[3], struct drand48_data *buffer, long int *result)
{
	if (result)
	{
		get_random_bytes((char*)result, sizeof(long int));
		*result = abs(*result);
		(*result) &= 0xffffffff;
		return 0;
	}

	return -1;
}

int mrand48_r(struct drand48_data *buffer,long int *result)
{
	if (result)
	{
		get_random_bytes((char*)result, sizeof(long int));
		return 0;
	}

	return -1;
}

int jrand48_r(unsigned short int xsubi[3], struct drand48_data *buffer, long int *result)
{
	if (result)
	{
		get_random_bytes((char*)result, sizeof(long int));
		return 0;
	}

	return -1;
}

int srand48_r(long int seedval, struct drand48_data *buffer)
{
#if PUT_BYTES == 1
	put_random_bytes((char *)&seedval, sizeof(seedval));
#endif

	return 0;
}

int seed48_r(unsigned short int seed16v[3], struct drand48_data *buffer)
{
#if PUT_BYTES == 1
	put_random_bytes((char *)&(seed16v[0]), sizeof(unsigned short));
	put_random_bytes((char *)&(seed16v[1]), sizeof(unsigned short));
	put_random_bytes((char *)&(seed16v[2]), sizeof(unsigned short));
#endif

	return 0;
}

int lcong48_r(unsigned short int param[7], struct drand48_data *buffer)
{
#if PUT_BYTES == 1
	int index;
	for(index=0; index<7; index++)
		put_random_bytes((char *)&param[index], sizeof(unsigned short));
#endif

	return 0;
}
